package com.ibm.bpac.baan.util;

// This utility daemon application is used to put an ePic header on a Baan XML 
// message and place it on an MQSeries Q by using MQAK API's
//
// Usage:
// java AddEpicHdrToBaanDaemon <sleeptime>
//************************************************************************************
import com.ibm.epic.adapters.eak.mcs.*;
import com.ibm.epic.adapters.eak.nativeadapter.*;
import com.ibm.epic.adapters.eak.common.*;
import com.ibm.epic.trace.exception.*;          // EpicTraceException
import com.ibm.epic.trace.client.*;			// EpicTraceClient
import com.ibm.logging.*;                       // IRecordType.TYPE_xxx

import java.util.*; 
import java.lang.*;
import java.io.*;

 /**
 * source file name: AddEpicHdrToBaanDaemon.java
 *
 * 
 *
 * @author IBM AIM Services
 * @version 1.00, 08 August 2001
 */ 

public class AddEpicHdrToBaanDaemon
{
	boolean stopNow = false;					// tells the Daemon to stop.
		
	int i, rc;
	long sleeptime;

	/** Default component Name */
	// private static String defaultComponentName = "DefaultComponentName";
	
	/** Application ID, from ADAPTER_AppID in aqmsetup file. 
	 * This value will be used to locate the section of the aqmconfig(.xml) file to use in determining
	 * what queue to put the XML messages to, and what epic header format to use (MQPP or MQRFH2).
	 */
	protected String epicAppID;
	
	/** Full path to directory to be used for processing.  
	 * This is shared with the Baan custom program.
	 * From BAANOUT_DIRECTORY in aqmsetup file
	 */
	protected String filepath;
	
	/** Filename used by Baan when it has output data.  
	 * Baan renames its own working file to this name when it is ready to pass it.
	 * Baan will also rename it back to an internal working name if it is still 
	 * there when more data needs to be passed. The file may contain multiple records,
	 * each record being a complete XML document.
	 * From BAAN_PASS_FILENAME in aqmsetup file
	 */
	protected String passedFileName;
	
	/** Filename to which the passed Baan file is renamed to 
	 * prior to this daemon doing any work with it.
	 * Thus, Baan may be putting out additional records while this daemon is processing the previous ones.
	 * From ADAPTER_WORKING_FILENAME in aqmsetup file
	 */
	protected String workingFileName;

	protected String baanFileName;	// filepath + passedFileName
	protected File baanFile;
	protected String workFileName;	// filepath + workingFileName
	protected File workFile;
	
	/** body type, set to root element name	 */
	protected String bodyType; 
	
	/** body category, set to Baan4 at present	 */
    protected String bodyCategory; 


	protected EpicNativeAdapter ENA = null;
	// Message type is needed but meaningless at present. Use a constant
	String msgType = "APPLICATION";
	
	/** component name, needed but meaningless. Constant used.	 */
    String componentName = "Component"; 

	
	protected boolean trace = false;
	protected EpicTraceClient traceClient = null;
	
	String delim;
//=========================================================================================
	/**
	* Empty constructor. Sets sleeptime to 2 seconds.  Sets BodyCategory to "Baan4".
	*/
	public AddEpicHdrToBaanDaemon()
	{
		sleeptime = 2000;
		bodyCategory = "Baan4";
		System.out.println("Starting up.");

	}
	/** One parm constructor. Sets sleeptime to parameter.  Sets BodyCategory to "Baan4".
	*/
	public AddEpicHdrToBaanDaemon(long _sleeptime)
	{
	    sleeptime = _sleeptime;
		bodyCategory = "Baan4";
		System.out.println("Starting up.");
		
	}
	/** Two parm constructor. Sets sleeptime and BodyCategory to passed parameters.
	 */
	public AddEpicHdrToBaanDaemon(long _sleeptime, String _bodyCategory)
	{
	    sleeptime = _sleeptime;
		bodyCategory = _bodyCategory;
		System.out.println("Starting up.");
	}


	/**
	* Checks for existence of the Baan flatfile.
	* For each XML document in the flatfile, puts an ePic hdr on it and
	* calls MQAK to place it on an MQSeries Q. Then sleeps a little and
	* repeats the process.
	* Keeps running until the user hits "<Enter>".
	* @param int _sleeptime
	*/
	public int runDaemon(long _sleeptime) throws Exception
	{
		Object [] newValues;
		Object [] tempObj;

		sleeptime = _sleeptime;
		// System.out.println("runDaemon started.");
		rc = 0;
		delim = File.separator;  // Get the correct delimiter for this machine
		
		// Read the aqmsetup file into memory and then extract:
		// ApplID, File path and file names to use
		try
		{
			AdapterCfg.readCfgFile("appCodeNm");
		}
		catch (Exception e)
		{
			System.out.println("ERROR! The aqmsetup file could not be located. " + 
							   "Check the AQMSETUPFILE environment variable.");
			return -1;
		}
//		try
//		{ 
			epicAppID = AdapterCfg.getKeyInfo("ADAPTER_AppID");
			if ( (epicAppID == null) | (epicAppID == "") )
			{
				System.out.println("ERROR! No value present in aqmsetup file " +
								   "for \"ADAPTER_AppID=\" ");
				return -1;
			}
			filepath = AdapterCfg.getKeyInfo("BAANOUT_DIRECTORY");
			if ( (filepath == null) || (filepath.length() == 0) )

			{
				System.out.println("ERROR! No value present in aqmsetup file " +
								   "for \"BAANOUT_DIRECTORY=\" "+
								   "(and current directory is NOT assumed for an empty path).");
				return -1;
			}
			passedFileName = AdapterCfg.getKeyInfo("BAAN_PASS_FILENAME");
			if ( (passedFileName == null) || (passedFileName.length() == 0) )
			{
				System.out.println("ERROR! No value present in aqmsetup file " +
								   "for \"BAAN_PASS_FILENAME=\" ");
				return -1;
			}
			workingFileName = AdapterCfg.getKeyInfo("ADAPTER_WORKING_FILENAME");
			if ( (workingFileName == null) || (workingFileName.length() == 0) )
			{
				workingFileName = "AddHdrWorkingFile.data";
				System.out.println("Warning! No value present in aqmsetup file " +
								   "for \"ADAPTER_WORKING_FILENAME=\". The default value of " +
								   workingFileName + " will be used.");
			}
			baanFileName = filepath + delim + passedFileName;
			workFileName = filepath + delim + workingFileName;
			
		try
		{ 
			// Now make sure the aqmconfig file is set up properly. 
			// NOTE: We are only checking that a section exists for the AppID in aqmsetup
		 	AdapterDirectory adapDir = new AdapterDirectory();
			String [] qryFilter = { "epicdestids" };           // [0]
			newValues = adapDir.getQueryResult("DEFAULT", "DEFAULT", epicAppID, qryFilter);
			tempObj = null;
			// NOTE: tempObj is also an array since the same tag could be repeated in aqmconfig.
			//       We will only use the first occurrence.
			
			// [0]: destids  -- this should be present
			tempObj = (Object[])newValues[0];
			if (null == tempObj[0])
			{
				System.out.println("ERROR! - There was a problem with the aqmconfig file. " +
							   "Check the aqmsetup file to see if it is pointing to the " +
							   "correct aqmconfig, and that the aqmconfig has a section for " +
							   "the ePic app ID = " + epicAppID +  "." );
				return -1;
			}

			// Next, check to see if directory path exists, so we don't waste effort.
			File fp = new File(filepath);
			if (fp.exists() == false)
			{
				System.out.println("ERROR! Filepath value in aqmsetup file, \"BAANOUT_DIRECTORY=" + filepath
								   + "\". This filepath does not exist. " + 
								   "\n Compare to the environment variable of the same name. " +
								   "The values should match except for the ending delimiter, " +
								   "which should be present on the environment variable " +
								   "but not on the aqmsetup value. ");
				return -1;
			}
		
			
			/** Next, check for existence of a prior "working" file.
			 * This may occur if the daemon had crashed previously.
			 */
			// System.out.println("Onetime check for file:" + workFileName);
			workFile = new File(workFileName);
			if (workFile.exists() == true)
			{
				System.out.println("Prior workfile " + workFileName + " exists. Will be processed first.");
				rc = processFile(workFile);
				if (rc < 0)
				{
					// An error occurred, and error message already put out. Exit.
					return rc;
				}
			}
			else
			{
				// System.out.println("File " + workFileName + " does not exist. This is good.");
			}
		}
		catch (java.io.IOException ex3)
		{
			System.out.println("ERROR! - A flatfile Open error occurred: " + ex3);
			rc = -1;
		}
		catch (AdapterException ae)
		{
			System.out.println("ERROR! - There was a problem with the aqmconfig file. " +
							   "Check the aqmsetup file to see if it is pointing to the " +
							   "correct aqmconfig, and that the aqmconfig has a section for " +
							   "the ePic app ID = " + epicAppID +  "." + 
							   "\n The following exception occurred : " + ae);
			rc = -1;
		}
		catch (Exception e)
		{
			System.out.println("ERROR! - An Exception occurred: " + e);
			return -1;
		}
		
		if (rc < 0)
			return rc;
		
		try
		{
				
			/** Start a separate thread which will say when to stop this method.
			 * Supply ptr to this class instance so the variable "stopNow" can be set by the killer thread.
			 */
			AddEpicHdrToBaanDaemonKiller killer = new AddEpicHdrToBaanDaemonKiller(this);  
			Thread killerThread = new Thread(killer);
			// System.out.println("About to start the killer thread");
			
			killerThread.start();
			// System.out.println("After starting the killer thread");
		
			// The loop can be stopped by the killerThread setting stopNow to true.
			while (stopNow == false)
			{
				// First, check for existence of Baan passed file
				// System.out.println("Looking for file:" + baanFileName);

				baanFile = new File(baanFileName);
				if (baanFile.exists() == true)
				{
					// Flatfile exists, Rename and process it.
					workFile = new File(workFileName);
					if (baanFile.renameTo(workFile)== true)
					{
						// System.out.println("File successfully renamed to: " + workFileName);
						rc = processFile(workFile);
						if (rc < 0)   // error condition which ends the loop
							return rc;
					}
					else
					{
						System.out.println("ERROR! File " + baanFileName + " could not be renamed to " + 
										   workFileName + ". This may be a permissions problem. " +
										   "\n Check and fix before restarting the daemon.");
						return -1;
					}
				}
				else
				{
					// System.out.println("Did not find file:" + baanFileName);
				}
				// Now sleep for a little while before looping again
				try 
				{
					Thread.sleep(sleeptime);
				}
				catch (InterruptedException ie) { /* skip it */ }
			}

			System.out.println("Shutting down.");
		}

		// If an error has occurred in the above, try to identify what went wrong

		// Was it an MQSeries error?
		catch (AdapterException ae)
		{
			System.out.println("An MQAK error occurred : " + ae);
			rc = -1;
		}
		// Was it a file open error?
		catch (java.io.IOException ex3)
		{
			System.out.println("A flatfile Open error occurred: " + ex3);
			rc = -1;
		}
		catch (Exception e)
		{
			System.out.println("An Exception occurred: " + e);
			rc = -1;
		}
		return rc;
	}
	
	//************************************************************************************
	
	/** Process all the XML messages on the flatfile. Delete the flatfile when done.
	 */
	private int processFile(File _thisFile) throws Exception
	{
		int my_rc = 0;
		int BUFFER_SIZE = 32000;
		String xmlMsg = null;
		// System.out.println("processFile method entered.");

		try
		{
			// Open the file and set up a loop to read the records
			FileReader fr = new FileReader(_thisFile);
			BufferedReader br = new BufferedReader(fr, BUFFER_SIZE);
			while ((xmlMsg = br.readLine()) != null)
			{
				// System.out.println("Message from Baan: \n" + xmlMsg);
				if (MQAOXMLIJ.isXML(xmlMsg) == false)
				{
				    System.out.println("Message from Baan: \n" + xmlMsg);
					System.out.println("Message is not valid XML and will be ignored.");
				}
				else
				{
					// Find the root element
					// System.out.println("\nParsing the XML String");
					MQAOCSRJ csr = new MQAOCSRJ(); // Cursors maintain a reference to your
												   // position within the parse tree.
					MQAOXMLIJ.parse (csr, xmlMsg); // Parse the document.
					MQAOXMLIJ.getFirstElement(csr); //Cursor now at root element
					
					bodyType = MQAOXMLIJ.getNameAtCursor(csr);  //This is it
					// System.out.println("bodyType <" + bodyType + ">");
					// System.out.println("bodyCategory <" + bodyCategory + ">");
					try
					{
						// Create an epic message (this is what the epicnativeadapter sends)
						EpicMessage epicMsgObj = new EpicMessage(epicAppID,msgType);
		
						//epicMsgObj.setDestinationLogicalID(epicAppID);  // EpicMessage Destination ID
						epicMsgObj.setBodyCategory(bodyCategory);  // EpicMessage Body Category
						epicMsgObj.setBodyType(bodyType);   // EpicMessage Body Type

						epicMsgObj.setBodyData(xmlMsg);      // This can be an XML document
						/** Create an EpicNativeAdapter to send the message.  */
						// componentName can be any string you want.
						ENA = new EpicNativeAdapter(epicAppID, componentName);
		
						// Send the message object
						ENA.sendMsg(epicMsgObj);
						System.out.println("Baan XML message put on the queue. BodyCategory <" + 
										   bodyCategory + ">, BodyType <" + bodyType + ">.");
					}
					catch (AdapterException e) 
					{
						System.out.println("Exception received from MQAK: " + e);
						// Kill the daemon
						stopNow = true;
						my_rc = -1;
					}
				} //end of valid XML processing for one record
			} //end of record loop
			
			// Done. Close the file and delete it.
			br.close();
			fr.close();
			_thisFile.delete();
			
		}
		catch (Exception e)
		{
			System.out.println("Unexpected Exception received: \n" + e);
			// Kill the daemon
			stopNow = true;
			my_rc = -1;
		}
		return my_rc;
	}


	//************************************************************************************
	//************************************************************************************

	public static void main(String args[]) throws Exception
	{
		int return_code = 0;
		long my_sleeptime = 2000;
		if(args.length > 1)
		{
			System.out.println("ERROR - wrong number of arguments.");
			System.out.println("Usage: java AddEpicHdrToBaanDaemon <sleeptime (optional)> ");
			return;
		}
		else if (args.length == 0)
			{
			System.out.println("Sleeptime not passed as an argument. Default of " + my_sleeptime 
						   + " (milliseconds) used.");
		}
		else
		{
			try
			{
				my_sleeptime = Long.parseLong(args[0]); 
			}
			catch (NumberFormatException ne)
			{
				System.out.println("Invalid Sleeptime passed as an argument. Default of " + my_sleeptime 
						   + " (milliseconds) used.");
			}
		}
	
	
		// AddEpicHdrToBaanDaemon myClass = new AddEpicHdrToBaanDaemon(my_sleeptime);
 		AddEpicHdrToBaanDaemon myClass = new AddEpicHdrToBaanDaemon();
		return_code = myClass.runDaemon(my_sleeptime);
		
		if (return_code < 0)
		{
			System.out.println("Please note and fix the above error conditions before restarting " + 
							   "this program. \n Press <Enter> to close. \n");
			BufferedReader br = new BufferedReader(new InputStreamReader(System.in));
			try
			{
				br.read(); // don't care what the user enters, just that he enters.
			}
			catch(IOException e) 
			{
			} // end catch code
		}


	}
} // end of AddEpicHdrToBaanDaemon

class AddEpicHdrToBaanDaemonKiller implements Runnable 
{
	AddEpicHdrToBaanDaemon addHdr;
	String s;

    public AddEpicHdrToBaanDaemonKiller(AddEpicHdrToBaanDaemon _addHdr) 
	{
		addHdr = _addHdr;
    }

    public void run() 
	{
		System.out.println("Press <Enter> to stop the AddEpicHdrToBaanDaemon. \n");
		BufferedReader br = new BufferedReader(new InputStreamReader(System.in));
		try
		{
			s = br.readLine(); // don't care what the user enters, just that he enters.
		}
		catch(IOException e) 
		{
		} // end catch code
		System.out.println("OK. The daemon will stop in a few seconds. \n");
		// Setting stopNow will end the every-10-seconds loop in AddEpicHdrToBaanDaemon
		addHdr.stopNow = true;
		
    }
}